<?php

/**
 * Returns a mongodb collection object.
 */
function _cachetags_mongodb_collection($collection_name) {
  global $db_prefix;
  static $mongo, $collections, $simpletest_prefix;
  // Allow specific DB for cachetags if specified.
  if (!$server_name = variable_get('cachetags_mongodb_connection', FALSE)) {
    // Otherwise use a unified connection under MongoDB module.
    require_once dirname(__FILE__) . '/../mongodb/mongodb.module';
    return mongodb_collection($collection_name);
  }
  if (!$db_name = variable_get('cachetags_mongodb_dbname', FALSE)) {
    $db_name = variable_get('mongodb_dbname', 'drupal');
  }

  // We call this function earlier than the database is initalized so we would
  // read the parent collection without this.
  if (!isset($simpletest_prefix)) {
    if (isset($_SERVER['HTTP_USER_AGENT']) && preg_match("/^(simpletest\d+);/", $_SERVER['HTTP_USER_AGENT'], $matches)) {
      $simpletest_prefix = $matches[1];
    }
    else {
      $simpletest_prefix = '';
    }
  }
  // However, once $db_prefix is filled by the database, the simpletest_prefix
  // is no longer needed.
  if ($simpletest_prefix && $db_prefix && strpos($db_prefix, $simpletest_prefix) !== FALSE) {
    $simpletest_prefix = '';
  }
  $collection_name = $db_prefix . $simpletest_prefix . $collection_name;
  if (!isset($mongo)) {
    $mongo = new Mongo($server_name);
    $mongo = $mongo->selectDB($db_name);
  }
  if (!isset($collections[$collection_name])) {
    $collections[$collection_name] = $mongo->selectCollection($collection_name);
  }
  return $collections[$collection_name];
}

/**
 * Return data from the persistent cache. Data may be stored as either plain text or as serialized data.
 * cache_get will automatically return unserialized objects and arrays.
 *
 * @param $cid
 *   The cache ID of the data to retrieve.
 * @param $table
 *   The table $table to store the data in. Valid core values are 'cache_filter',
 *   'cache_menu', 'cache_page', or 'cache' for the default cache.
 */
function cache_get($cid, $table = 'cache') {
  global $user;

  try {
    $collection = _cachetags_mongodb_collection($table);

    // Garbage collection necessary when enforcing a minimum cache lifetime.
    $cache_flush = variable_get('cache_flush_' . $table, 0);
    if ($cache_flush && ($cache_flush + variable_get('cache_lifetime', 0) <= $_SERVER['REQUEST_TIME'])) {
      // Reset the variable immediately to prevent a meltdown in heavy load situations.
      variable_set('cache_flush_' . $table, 0);
      // Time to flush old cache data

      $find = array('expire' => array('$lte' => $cache_flush, '$ne' => CACHE_PERMANENT));
      $collection->remove($find);
    }

    $cache = $collection->findOne(array('_id' => (string)$cid));
  }
  catch (Exception $e) {
    // Avoid exception crashes and just return nothing.
    return FALSE;
  }

  if (!$cache || !isset($cache['data'])) {
    return FALSE;
  }
  unset($cache['_id']);
  $cache = (object)$cache;
  // If enforcing a minimum cache lifetime, validate that the data is
  // currently valid for this user before we return it by making sure the cache
  // entry was created before the timestamp in the current session's cache
  // timer. The cache variable is loaded into the $user object by _drupal_session_read()
  // in session.inc. If the data is permanent or we're not enforcing a minimum
  // cache lifetime always return the cached data.
  if ($cache->expire != CACHE_PERMANENT && variable_get('cache_lifetime', 0) && $user->cache > $cache->created) {
    // This cache data is too old and thus not valid for us, ignore it.
    return FALSE;
  }

  // The cache data is invalid if any of its tags have been cleared since.
  if ($cache->tags && cache_tags_is_expired($cache->checksum, $cache->tags)) {
    return FALSE;
  }

  if ($cache->data instanceof MongoBinData) {
    $cache->data = $cache->data->bin;
  }

  if ($cache->serialized) {
    $cache->data = unserialize($cache->data);
  }
  return $cache;
}

/**
 * Store data in the persistent cache.
 *
 * The persistent cache is split up into four database
 * tables. Contributed modules can add additional tables.
 *
 * 'cache_page': This table stores generated pages for anonymous
 * users. This is the only table affected by the page cache setting on
 * the administrator panel.
 *
 * 'cache_menu': Stores the cachable part of the users' menus.
 *
 * 'cache_filter': Stores filtered pieces of content. This table is
 * periodically cleared of stale entries by cron.
 *
 * 'cache': Generic cache storage table.
 *
 * The reasons for having several tables are as follows:
 *
 * - smaller tables allow for faster selects and inserts
 * - we try to put fast changing cache items and rather static
 *   ones into different tables. The effect is that only the fast
 *   changing tables will need a lot of writes to disk. The more
 *   static tables will also be better cachable with MySQL's query cache
 *
 * @param $cid
 *   The cache ID of the data to store.
 * @param $data
 *   The data to store in the cache. Complex data types will be automatically serialized before insertion.
 *   Strings will be stored as plain text and not serialized.
 * @param $table
 *   The table $table to store the data in. Valid core values are 'cache_filter',
 *   'cache_menu', 'cache_page', or 'cache'.
 * @param $expire
 *   One of the following values:
 *   - CACHE_PERMANENT: Indicates that the item should never be removed unless
 *     explicitly told to using cache_clear_all() with a cache ID.
 *   - CACHE_TEMPORARY: Indicates that the item should be removed at the next
 *     general cache wipe.
 *   - A Unix timestamp: Indicates that the item should be kept at least until
 *     the given time, after which it behaves like CACHE_TEMPORARY.
 * @param $headers
 *   A string containing HTTP header information for cached pages.
 * @param $tags
 *   An array of tags to be stored with the cache item. These should normally
 *   identify objects used to build the cache item, which should trigger
 *   cache invalidation when updated. For example if a cached item represents
 *   a node, both the node ID and the author's user ID might be passed in as
 *   tags. For example array('node' => array(123), 'user' => array(92)).
 */
function cache_set($cid, $data, $table = 'cache', $expire = CACHE_PERMANENT, $headers = NULL, $tags = array()) {
  $scalar = is_scalar($data);
  $entry = array(
    '_id' => (string)$cid,
    'cid' => (string)$cid,
    'created' => $_SERVER['REQUEST_TIME'],
    'expire' => $expire,
    'serialized' => !$scalar,
    'headers' => $headers,
    'data' => $scalar ? $data : serialize($data),
    'tags' => $tags,
    'checksum' => cache_tags_checksum($tags),
  );

  try {
    // Use MongoBinData for non-UTF8 strings.
    if (is_string($entry['data']) && !drupal_validate_utf8($entry['data'])) {
      $entry['data'] = new MongoBinData($entry['data']);
    }
    $collection = _cachetags_mongodb_collection($table);
    $collection->save($entry);
  }
  catch (Exception $e) {
    // The database may not be available, so we'll ignore cache_set requests.
  }
}

/**
 *
 * Expire data from the cache. If called without arguments, expirable
 * entries will be cleared from the cache_page and cache_block tables.
 *
 * @param $cid
 *   If set, the cache ID to delete. Otherwise, all cache entries that can
 *   expire are deleted.
 *
 * @param $table
 *   If set, the table $table to delete from. Mandatory
 *   argument if $cid is set.
 *
 * @param $wildcard
 *   If $wildcard is TRUE, cache IDs starting with $cid are deleted in
 *   addition to the exact cache ID specified by $cid.  If $wildcard is
 *   TRUE and $cid is '*' then the entire table $table is emptied.
 */
function cache_clear_all($cid = NULL, $table = NULL, $wildcard = FALSE) {
  global $user;

  if (!isset($cid) && !isset($table)) {
    // Clear the block cache first, so stale data will
    // not end up in the page cache.
    cache_clear_all(NULL, 'cache_block');
    cache_clear_all(NULL, 'cache_page');
    return;
  }

  try {
    $collection = _cachetags_mongodb_collection($table);

    if (empty($cid)) {
      $now = time();
      if (variable_get('cache_lifetime', 0)) {
        // We store the time in the current user's $user->cache variable which
        // will be saved into the sessions table by sess_write(). We then
        // simulate that the cache was flushed for this user by not returning
        // cached data that was cached before the timestamp.
        $user->cache = $now;

        $cache_flush = variable_get('cache_flush_'. $table, 0);
        if ($cache_flush == 0) {
          // This is the first request to clear the cache, start a timer.
          variable_set('cache_flush_'. $table, $now);
        }
        else if (time() > ($cache_flush + variable_get('cache_lifetime', 0))) {
          // Clear the cache for everyone, cache_lifetime seconds have
          // passed since the first request to clear the cache.
          $collection->remove(array('expire' => array('$ne' => CACHE_PERMANENT, '$lte' => $now)));
          variable_set('cache_flush_'. $table, 0);
        }
      }
      else {
        // No minimum cache lifetime, flush all temporary cache entries now.
        $collection->remove(array('expire' => array('$ne' => CACHE_PERMANENT, '$lte' => $now)));
      }
    }
    else {
      if ($wildcard) {
        if ($cid == '*') {
          $collection->remove();
        }
        else {
          $collection->remove(array('_id' => new MongoRegex('/'. preg_quote($cid, '/') .'/')));
        }
      }
      else {
        $collection->remove(array('_id' => (string)$cid));
      }
    }
  }
  catch (Exception $e) {
    // Ignore exceptions.
  }
}

function &cache_tags_cache() {
  static $cache = array();
  return $cache;
}

function cache_tags_checksum($tags) {
  $checksum = 0;
  $query_tags = array();
  $tag_cache = &cache_tags_cache();

  foreach (cache_tags_flatten($tags) as $tag) {
    if (isset($tag_cache[$tag])) {
      $checksum += $tag_cache[$tag];
    }
    else {
      $query_tags[] = $tag;
    }
  }
  if ($query_tags) {
    try {
      foreach (_cachetags_mongodb_collection('cache_tags')->find(array('_id' => array('$in' => $query_tags))) as $db_tag) {
        $tag_cache[$db_tag['_id']] = $db_tag['invalidations'];
        $checksum += $db_tag['invalidations'];
      }
    }
    catch (Exception $e) {

    }
  }

  return $checksum;
}

function cache_tags_flatten($tags) {
  if (isset($tags[0])) return $tags;
  $flat_tags = array();
  foreach ($tags as $namespace => $values) {
    if (is_array($values)) {
      foreach ($values as $value) {
        $flat_tags[] = "$namespace:$value";
      }
    }
    else {
      $flat_tags[] = "$namespace:$values";
    }
  }
  return $flat_tags;
}

function cache_tags_is_expired($checksum, $tags) {
  return $checksum < cache_tags_checksum($tags);
}

function cache_invalidate($tags) {
  $tag_cache = &cache_tags_cache();
  try {
    foreach (cache_tags_flatten($tags) as $tag) {
      unset($tag_cache[$tag]);
      _cachetags_mongodb_collection('cache_tags')->update(array('_id' => $tag), array('$inc' => array('invalidations' => 1)), array('upsert' => TRUE));
    }
  }
  catch (Exception $e) {
    // Ignore exceptions.
  }
}
